==> acst-0.1.1.tar.gz: acst-0.1.1/acst.c <==
/**
 * Copyright (C) 2022 Chris Noxz
 * Author(s): Chris Noxz <chris@noxz.tech>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include <dirent.h>            // for dirent, closedir, opendir, readdir, DIR
#include <errno.h>             // for errno, ENODATA
#include <fcntl.h>             // for open, O_NOFOLLOW, O_RDONLY
#include <stdarg.h>            // for va_end, va_list, va_start
#include <stdbool.h>           // for true, false, bool
#include <stdio.h>             // for fprintf, stdout, snprintf, sprintf
#include <stdlib.h>            // for free, EXIT_FAILURE, calloc, exit, malloc
#include <string.h>            // for strlen, strcmp, memset, strerror
#include <time.h>              // for timespec
#include <unistd.h>            // for close, read
#include <bits/getopt_core.h>  // for getopt, optind
#include <sys/stat.h>          // for stat, fstat, lstat, S_ISDIR, S_ISREG
#include <sys/xattr.h>         // for fgetxattr, fremovexattr, fsetxattr

#include "sha256.h"            // for SHA256_final, SHA256_init, SHA256_update
#include "acst.h"


static void
bin2hex(unsigned char *bin, size_t len, char *hex)
{
	unsigned int    i,                      // index counter of binary data
	                j;                      // index counter of hex data
	for (i = j = 0; i < len; i++) {
		hex[j++] = hextab[bin[i] >> 4];
		hex[j++] = hextab[bin[i] & 0x0F];
	}
	hex[j] = 0;
}


static void
error(enum Error er, const char *fmt, ...)
{
	va_list         ap;                     // argument list holder

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	switch (er) {
	case ER_NOT_REGULAR:                    cnt.errNotRegular++; break;
	case ER_OPENING:                        cnt.errOpening++; break;
	case ER_XATTR_OPERATION:                cnt.errWritingXattr++; break;
	case ER_GENERIC:                        cnt.errGeneric++; break;
	case ER_FATAL:                          exit(EXIT_FAILURE);
	}
	cnt.errs++;
}


static enum FileState
file_state(int fd, xa_t *xa_s, xa_t *xa_a)
{
	if (arg.remove)
		return xa_write(fd, NULL) ? FS_REMOVED : FS_REMOVED_ERROR;

	if (!xa_compute(fd, xa_a))
		return FS_ERROR;
	else if (xa_a->cs[0] == 0)
		return FS_DISRUPTED;

	if (!xa_read(fd, xa_s))
		return FS_MALFORMED;

	if (xa_s->cs[0] == 0)
		return FS_NEW;

	/* compare stored and actual timestamp */
	xa_a->tcmp = (
		  (int)(xa_s->s - xa_a->s) > 0 || (int)(xa_s->ns - xa_a->ns) > 0 ? 1
		: (int)(xa_s->s - xa_a->s) < 0 || (int)(xa_s->ns - xa_a->ns) < 0 ? -1
		: 0
	);

	return (
		  strcmp(xa_s->cs, xa_a->cs) == 0   // if hash is correct
		? xa_a->tcmp == 0 ? FS_OK           // ...and ts are matching
		: FS_TOUCHED                        // ...or if not
		: xa_a->tcmp < 0 ? FS_OUTDATED      // else if ts is new
		: xa_a->tcmp > 0 ? FS_BACKDATED     // else if ts is old
		: FS_CORRUPT                        // else
	);
}


static void
process_argument(const char *fn)
{
	struct stat     st;                     // stat holder

	if (lstat(fn, &st) != 0)
		error(ER_OPENING, STR_ERR_OPNF, fn);
	else if (S_ISREG(st.st_mode))
		process_file(fn);
	else if (S_ISDIR(st.st_mode))
		process_directory(fn);
	/* make sure to only process regular files or directory walking */
	else
		error(ER_NOT_REGULAR, STR_ERR_REGF, fn);
}


static void
process_directory(const char *fn)
{
	DIR             *d;                     // directory pointer
	struct dirent   *dp;                    // directory entry pointer
	struct stat     dst,                    // stat holder for directory
	                st;                     // stat holder for children
	char            *rp;                    // relative path holder

	if (!arg.recursive)
		return error(ER_NOT_REGULAR, STR_ERR_RECU, fn);

	/* open directory and get stats */
	if (!((d = opendir(fn)) && lstat(fn, &dst) == 0))
		return error(ER_OPENING, STR_ERR_OPND, fn);

	/* loop through directory (including hidden files), making sure to process
	 * the relative path of each file */
	while ((dp = readdir(d)) != NULL) {
		if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0)
			continue;
		if ((rp = malloc(strlen(fn) + strlen(dp->d_name) + 2))) {
			sprintf(rp, "%s/%s", fn, dp->d_name);
			/* make sure not to descend directories on other filesystems */
			if (!(lstat(rp, &st) == 0 && dst.st_dev == st.st_dev))
				error(ER_OPENING, STR_ERR_DDEV, rp, fn);
			else
				process_argument(rp);
			free(rp);
		}
	}
	closedir(d);
}


static void
process_file(const char *fn)
{
	int             fd;                     // file holder
	xa_t            xa_s,                   // stored attribute
	                xa_a;                   // actual attribute
	enum FileState  st;                     // file state

	if (!(fd = open(fn, O_RDONLY | O_NOFOLLOW)))
		return error(ER_OPENING, STR_ERR_OPNF, fn);

	if ((st = file_state(fd, &xa_s, &xa_a)) == FS_ERROR)
		error(ER_GENERIC, STR_ERR_HASH, fn);
	else if (st == FS_REMOVED_ERROR)
		error(ER_XATTR_OPERATION, STR_ERR_XARM, fn, strerror(errno));
	else {
		report(fn, st);
		if (!arg.quiet && !(st == FS_OK || st == FS_REMOVED || st == FS_DISRUPTED || st == FS_MALFORMED))
			fprintf(stdout, STR_CMP_FRMT
				, strlen(xa_s.cs) > 0 ? xa_s.cs : zSHA256
				, xa_s.s
				, xa_s.ns
				, strlen(xa_a.cs) > 0 ? xa_a.cs : zSHA256
				, xa_a.s
				, xa_a.ns
			);
		if (!(st == FS_OK || st == FS_REMOVED || st == FS_DISRUPTED || st == FS_CORRUPT || st == FS_BACKDATED || st == FS_MALFORMED) && !xa_write(fd, &xa_a))
			error(ER_XATTR_OPERATION, STR_ERR_XAWR, fn);
	}

	close(fd);
}


static void
report(const char *fn, enum FileState st)
{
	switch (st) {
	case FS_OK:
	case FS_REMOVED:                        cnt.ok++; break;
	case FS_DISRUPTED:                      cnt.disrupted++; break;
	case FS_CORRUPT:                        cnt.corrupt++; break;
	case FS_TOUCHED:                        cnt.touched++; break;
	case FS_NEW:                            cnt.new++; break;
	case FS_OUTDATED:                       cnt.outdated++; break;
	case FS_BACKDATED:                      cnt.backdated++; break;
	case FS_MALFORMED:                      cnt.malformed++; break;
	/* don't report errors here (this is only to avoid compiler warnings) */
	case FS_REMOVED_ERROR:
	case FS_ERROR:                          return;
	}
	cnt.total++;

	if (st == FS_CORRUPT || st == FS_BACKDATED || st == FS_MALFORMED) {
		fprintf(stderr, STR_ERR_ABNO, fn);
		fprintf(stdout, STR_OUT_STAT
			, st == FS_CORRUPT ? STR_OUT_CORR
			: st == FS_BACKDATED ? STR_OUT_BACK
			: STR_OUT_MALF, fn);
	}

	if (arg.quiet < 2 && (st == FS_DISRUPTED || st == FS_TOUCHED || st == FS_NEW || st == FS_OUTDATED))
		fprintf(stdout, STR_OUT_STAT
			, st == FS_DISRUPTED ? STR_OUT_DISR
			: st == FS_TOUCHED ? STR_OUT_TOUC
			: st == FS_NEW ? STR_OUT_NEW
			: STR_OUT_OUTD, fn);

	if (arg.quiet < 1 && (st == FS_OK || st == FS_REMOVED))
		fprintf(stdout, STR_OUT_STAT
			, st == FS_OK ? STR_OUT_OK
			: STR_OUT_REMO, fn);
}


static bool
xa_compute(int fd, xa_t *xa)
{
	struct stat     st;                     // file stat holder

	if ((fstat(fd, &st) != 0))
		return false;

	/* retrieve date modified and compute hash of file */
	xa->s = st.st_mtim.tv_sec;
	xa->ns = st.st_mtim.tv_nsec;
	xa_hash(fd, xa->cs);

	/* get file stats after hash computation as the file could have been
	 * changed or deleted during the computation */
	if ((fstat(fd, &st) != 0))
		return false;

	/* check if the file was modified while the hash was computed, and of so
	 * indicate with an empty hash */
	if (!((long unsigned int)st.st_mtim.tv_sec == xa->s && (long unsigned int)st.st_mtim.tv_nsec == xa->ns))
		xa->cs[0] = 0;

	return true;
}


static void
xa_hash(int fd, char *cs)
{
	SHA256_ctx      ctx;                    // SHA256 context
	unsigned char   *hash,                  // hash buffer
	                buf[BUFSZ];             // file content read buffer
	size_t          len;                    // holder for read buffer length

	SHA256_init(&ctx);
	while ((len = read(fd, buf, BUFSZ)))
		SHA256_update(&ctx, buf, len);

	if (!(hash = calloc(1, SHA256_BYTES)))
		error(ER_FATAL, STR_ERR_OOM);

	SHA256_final(&ctx, hash);

	bin2hex(hash, SHA256_BYTES, cs);
	free(hash);
}


static bool
xa_read(int fd, xa_t *xa)
{
	char            *c;                     // character pointer used in loop

	/* initialize metadata holders to zero, and create a blank xa */
	memset(xa, 0, sizeof(*xa));
	memset(ts, 0, sizeof(*ts));

	/* read checksum, on error indicate as new if missing else fail */
	if (fgetxattr(fd, XATTR_CS, xa->cs, sizeof(xa->cs)) < 0)
		return errno == ENODATA ? true : false;
	else if (strlen(xa->cs) != CSSZ)
		return false;

	/* make sure digits are hex (else fail) and in lower case (else adjust) */
	for (c = xa->cs; *c; c++)
		if (*c >= 0x41 && *c <= 0x46)
			*c += 0x20;
		else if ((*c < 0x30 || *c > 0x39) && (*c < 0x61 || *c > 0x66))
			return false;

	/* read timestamp, on error (ENODATA included) fail */
	if (fgetxattr(fd, XATTR_TS, ts, sizeof(ts) - 1) < 0)
		return false;
	else if (sscanf(ts, FRMT_TS, &xa->s, &xa->ns) < 1)
		return false;

	return true;
}


static bool
xa_write(int fd, const xa_t *xa)
{
	if (arg.dryrun)
		return true;

	/* remove extended attributes from file if xa is null */
	if (xa == NULL)
		return (
			fremovexattr(fd, XATTR_TS) |
			fremovexattr(fd, XATTR_CS)
		) == 0;

	/* prepare and write extended attributes to file */
	snprintf(ts, sizeof(ts), FRMT_TS, xa->s, xa->ns);
	return (
		fsetxattr(fd, XATTR_TS, ts, strlen(ts), 0) |
		fsetxattr(fd, XATTR_CS, &xa->cs, CSSZ, 0)
	) == 0;
}


int
main(int argc, char *argv[])
{
	int             opt;                    // argument option holder

	arg.prg = argv[0];
	while ((opt = getopt(argc, argv, "hmnqrvx")) != -1) {
		switch (opt) {
		case 'h':                           USAGE(EXIT_SUCCESS); break;
		case 'm':                           arg.summarize = true; break;
		case 'n':                           arg.dryrun = true; break;
		case 'q':                           arg.quiet++; break;
		case 'r':                           arg.recursive = true; break;
		case 'v':                           VER(); break;
		case 'x':                           arg.remove = true; break;
		default:                            USAGE(EXIT_FAILURE);
		}
	}

	if (optind >= argc)
		USAGE(EXIT_FAILURE);

	for (int i = optind; i < argc; i++)
		process_argument(argv[i]);

	if (arg.summarize)
		SUMMERIZE();

	if ((cnt.corrupt + cnt.malformed + cnt.backdated) > 0)
		return 5;
	if (cnt.errs == 0 && (cnt.ok + cnt.outdated + cnt.touched + cnt.new) == cnt.total)
		return EXIT_SUCCESS;
	if (cnt.errOpening == cnt.errs)
		return 2;
	if (cnt.errNotRegular == cnt.errs)
		return 3;
	if (cnt.errWritingXattr == cnt.errs)
		return 4;
	return 6;
}

==> acst-0.1.2.tar.gz: acst-0.1.2/acst.c <==
/**
 * Copyright (C) 2022 Chris Noxz
 * Author(s): Chris Noxz <chris@noxz.tech>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include <dirent.h>            // for dirent, closedir, opendir, readdir, DIR
#include <errno.h>             // for errno, ENODATA
#include <fcntl.h>             // for open, O_NOFOLLOW, O_RDONLY
#include <stdarg.h>            // for va_end, va_list, va_start
#include <stdbool.h>           // for true, false, bool
#include <stdio.h>             // for fprintf, stdout, snprintf, sprintf
#include <stdlib.h>            // for free, EXIT_FAILURE, calloc, exit, malloc
#include <string.h>            // for strlen, strcmp, memset, strerror
#include <time.h>              // for timespec
#include <unistd.h>            // for close, read
#include <bits/getopt_core.h>  // for getopt, optind
#include <sys/stat.h>          // for stat, fstat, lstat, S_ISDIR, S_ISREG
#include <sys/xattr.h>         // for fgetxattr, fremovexattr, fsetxattr

#include "sha256.h"            // for SHA256_final, SHA256_init, SHA256_update
#include "acst.h"


static void
bin2hex(unsigned char *bin, size_t len, char *hex)
{
	unsigned int    i,                      // index counter of binary data
	                j;                      // index counter of hex data
	for (i = j = 0; i < len; i++) {
		hex[j++] = hextab[bin[i] >> 4];
		hex[j++] = hextab[bin[i] & 0x0F];
	}
	hex[j] = 0;
}


static void
error(enum Error er, const char *fmt, ...)
{
	va_list         ap;                     // argument list holder

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	switch (er) {
	case ER_NOT_REGULAR:                    cnt.errNotRegular++; break;
	case ER_OPENING:                        cnt.errOpening++; break;
	case ER_XATTR_OPERATION:                cnt.errWritingXattr++; break;
	case ER_GENERIC:                        cnt.errGeneric++; break;
	case ER_FATAL:                          exit(EXIT_FAILURE);
	}
	cnt.errs++;
}


static enum FileState
file_state(int fd, xa_t *xa_s, xa_t *xa_a)
{
	if (arg.remove)
		return xa_write(fd, NULL) ? FS_REMOVED : FS_REMOVED_ERROR;

	if (!xa_compute(fd, xa_a))
		return FS_ERROR;
	else if (xa_a->cs[0] == 0)
		return FS_DISRUPTED;

	if (!xa_read(fd, xa_s))
		return FS_MALFORMED;

	if (xa_s->cs[0] == 0)
		return FS_NEW;

	/* compare stored and actual timestamp */
	xa_a->tcmp = (
		  (int)(xa_s->s - xa_a->s) > 0 || (int)(xa_s->ns - xa_a->ns) > 0 ? 1
		: (int)(xa_s->s - xa_a->s) < 0 || (int)(xa_s->ns - xa_a->ns) < 0 ? -1
		: 0
	);

	return (
		  strcmp(xa_s->cs, xa_a->cs) == 0   // if hash is correct
		? xa_a->tcmp == 0 ? FS_OK           // ...and ts are matching
		: FS_TOUCHED                        // ...or if not
		: xa_a->tcmp < 0 ? FS_OUTDATED      // else if ts is new
		: xa_a->tcmp > 0 ? FS_BACKDATED     // else if ts is old
		: FS_CORRUPT                        // else
	);
}


static void
process_argument(const char *fn)
{
	struct stat     st;                     // stat holder

	if (lstat(fn, &st) != 0)
		error(ER_OPENING, STR_ERR_OPNF, fn);
	else if (S_ISREG(st.st_mode))
		process_file(fn);
	else if (S_ISDIR(st.st_mode))
		process_directory(fn);
	/* make sure to only process regular files or directory walking */
	else
		error(ER_NOT_REGULAR, STR_ERR_REGF, fn);
}


static void
process_directory(const char *fn)
{
	DIR             *d;                     // directory pointer
	struct dirent   *dp;                    // directory entry pointer
	struct stat     dst,                    // stat holder for directory
	                st;                     // stat holder for children
	char            *rp;                    // relative path holder

	if (!arg.recursive)
		return error(ER_NOT_REGULAR, STR_ERR_RECU, fn);

	/* open directory and get stats */
	if (!((d = opendir(fn)) && lstat(fn, &dst) == 0))
		return error(ER_OPENING, STR_ERR_OPND, fn);

	/* loop through directory (including hidden files), making sure to process
	 * the relative path of each file */
	while ((dp = readdir(d)) != NULL) {
		if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0)
			continue;
		if ((rp = malloc(strlen(fn) + strlen(dp->d_name) + 2))) {
			sprintf(rp, "%s/%s", fn, dp->d_name);
			/* make sure not to descend directories on other filesystems */
			if (!(lstat(rp, &st) == 0 && dst.st_dev == st.st_dev))
				error(ER_OPENING, STR_ERR_DDEV, rp, fn);
			else
				process_argument(rp);
			free(rp);
		}
	}
	closedir(d);
}


static void
process_file(const char *fn)
{
	int             fd;                     // file holder
	xa_t            xa_s,                   // stored attribute
	                xa_a;                   // actual attribute
	enum FileState  st;                     // file state

	if (!(fd = open(fn, O_RDONLY | O_NOFOLLOW)))
		return error(ER_OPENING, STR_ERR_OPNF, fn);

	if ((st = file_state(fd, &xa_s, &xa_a)) == FS_ERROR)
		error(ER_GENERIC, STR_ERR_HASH, fn);
	else if (st == FS_REMOVED_ERROR)
		error(ER_XATTR_OPERATION, STR_ERR_XARM, fn, strerror(errno));
	else {
		report(fn, st);
		if (!arg.quiet && !(st == FS_OK || st == FS_REMOVED || st == FS_DISRUPTED || st == FS_MALFORMED))
			fprintf(stdout, STR_CMP_FRMT
				, strlen(xa_s.cs) > 0 ? xa_s.cs : zSHA256
				, xa_s.s
				, xa_s.ns
				, strlen(xa_a.cs) > 0 ? xa_a.cs : zSHA256
				, xa_a.s
				, xa_a.ns
			);
		if (!(st == FS_OK || st == FS_REMOVED || st == FS_DISRUPTED || st == FS_CORRUPT || st == FS_BACKDATED || st == FS_MALFORMED) && !xa_write(fd, &xa_a))
			error(ER_XATTR_OPERATION, STR_ERR_XAWR, fn);
	}

	close(fd);
}


static void
report(const char *fn, enum FileState st)
{
	switch (st) {
	case FS_OK:
	case FS_REMOVED:                        cnt.ok++; break;
	case FS_DISRUPTED:                      cnt.disrupted++; break;
	case FS_CORRUPT:                        cnt.corrupt++; break;
	case FS_TOUCHED:                        cnt.touched++; break;
	case FS_NEW:                            cnt.new++; break;
	case FS_OUTDATED:                       cnt.outdated++; break;
	case FS_BACKDATED:                      cnt.backdated++; break;
	case FS_MALFORMED:                      cnt.malformed++; break;
	/* don't report errors here (this is only to avoid compiler warnings) */
	case FS_REMOVED_ERROR:
	case FS_ERROR:                          return;
	}
	cnt.total++;

	if (st == FS_CORRUPT || st == FS_BACKDATED || st == FS_MALFORMED) {
		fprintf(stderr, STR_ERR_ABNO, fn);
		fprintf(stdout, STR_OUT_STAT
			, st == FS_CORRUPT ? STR_OUT_CORR
			: st == FS_BACKDATED ? STR_OUT_BACK
			: STR_OUT_MALF, fn);
	}

	if (arg.quiet < 2 && (st == FS_DISRUPTED || st == FS_TOUCHED || st == FS_NEW || st == FS_OUTDATED))
		fprintf(stdout, STR_OUT_STAT
			, st == FS_DISRUPTED ? STR_OUT_DISR
			: st == FS_TOUCHED ? STR_OUT_TOUC
			: st == FS_NEW ? STR_OUT_NEW
			: STR_OUT_OUTD, fn);

	if (arg.quiet < 1 && (st == FS_OK || st == FS_REMOVED))
		fprintf(stdout, STR_OUT_STAT
			, st == FS_OK ? STR_OUT_OK
			: STR_OUT_REMO, fn);
}


static bool
xa_compute(int fd, xa_t *xa)
{
	struct stat     st;                     // file stat holder

	if ((fstat(fd, &st) != 0))
		return false;

	/* retrieve date modified and compute hash of file */
	xa->s = st.st_mtim.tv_sec;
	xa->ns = st.st_mtim.tv_nsec;
	xa_hash(fd, xa->cs);

	/* get file stats after hash computation as the file could have been
	 * changed or deleted during the computation */
	if ((fstat(fd, &st) != 0))
		return false;

	/* check if the file was modified while the hash was computed, and of so
	 * indicate with an empty hash */
	if (!((long unsigned int)st.st_mtim.tv_sec == xa->s && (long unsigned int)st.st_mtim.tv_nsec == xa->ns))
		xa->cs[0] = 0;

	return true;
}


static void
xa_hash(int fd, char *cs)
{
	struct SHA256_ctx ctx;                  // SHA256 context
	unsigned char   *hash,                  // hash buffer
	                buf[BUFSZ];             // file content read buffer
	size_t          len;                    // holder for read buffer length

	SHA256_init(&ctx);
	while ((len = read(fd, buf, BUFSZ)))
		SHA256_update(buf, len, &ctx);

	if (!(hash = calloc(1, SHA256_BYTES)))
		error(ER_FATAL, STR_ERR_OOM);

	SHA256_final(&ctx, hash);

	bin2hex(hash, SHA256_BYTES, cs);
	free(hash);
}


static bool
xa_read(int fd, xa_t *xa)
{
	char            *c;                     // character pointer used in loop

	/* initialize metadata holders to zero, and create a blank xa */
	memset(xa, 0, sizeof(*xa));
	memset(ts, 0, sizeof(*ts));

	/* read checksum, on error indicate as new if missing else fail */
	if (fgetxattr(fd, XATTR_CS, xa->cs, sizeof(xa->cs)) < 0)
		return errno == ENODATA ? true : false;
	else if (strlen(xa->cs) != CSSZ)
		return false;

	/* make sure digits are hex (else fail) and in lower case (else adjust) */
	for (c = xa->cs; *c; c++)
		if (*c >= 0x41 && *c <= 0x46)
			*c += 0x20;
		else if ((*c < 0x30 || *c > 0x39) && (*c < 0x61 || *c > 0x66))
			return false;

	/* read timestamp, on error (ENODATA included) fail */
	if (fgetxattr(fd, XATTR_TS, ts, sizeof(ts) - 1) < 0)
		return false;
	else if (sscanf(ts, FRMT_TS, &xa->s, &xa->ns) < 1)
		return false;

	return true;
}


static bool
xa_write(int fd, const xa_t *xa)
{
	if (arg.dryrun)
		return true;

	/* remove extended attributes from file if xa is null */
	if (xa == NULL)
		return (
			fremovexattr(fd, XATTR_TS) |
			fremovexattr(fd, XATTR_CS)
		) == 0;

	/* prepare and write extended attributes to file */
	snprintf(ts, sizeof(ts), FRMT_TS, xa->s, xa->ns);
	return (
		fsetxattr(fd, XATTR_TS, ts, strlen(ts), 0) |
		fsetxattr(fd, XATTR_CS, &xa->cs, CSSZ, 0)
	) == 0;
}


int
main(int argc, char *argv[])
{
	int             opt;                    // argument option holder

	arg.prg = argv[0];
	while ((opt = getopt(argc, argv, "hmnqrvx")) != -1) {
		switch (opt) {
		case 'h':                           USAGE(EXIT_SUCCESS); break;
		case 'm':                           arg.summarize = true; break;
		case 'n':                           arg.dryrun = true; break;
		case 'q':                           arg.quiet++; break;
		case 'r':                           arg.recursive = true; break;
		case 'v':                           VER(); break;
		case 'x':                           arg.remove = true; break;
		default:                            USAGE(EXIT_FAILURE);
		}
	}

	if (optind >= argc)
		USAGE(EXIT_FAILURE);

	for (int i = optind; i < argc; i++)
		process_argument(argv[i]);

	if (arg.summarize)
		SUMMERIZE();

	if ((cnt.corrupt + cnt.malformed + cnt.backdated) > 0)
		return 5;
	if (cnt.errs == 0 && (cnt.ok + cnt.outdated + cnt.touched + cnt.new) == cnt.total)
		return EXIT_SUCCESS;
	if (cnt.errOpening == cnt.errs)
		return 2;
	if (cnt.errNotRegular == cnt.errs)
		return 3;
	if (cnt.errWritingXattr == cnt.errs)
		return 4;
	return 6;
}
